// Parser.cpp: implementation of the CParser class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Token.h"
#include "Parser.h"

using namespace std;
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
bool CCommentParser::Parse( CScanner* pScanner)
{
	return pScanner->GetNextStringMatch("-->");
}

//
// Function name:bool CPIParser::Parse( CScanner* pScanner)
//
// Parse the processing instructionl.
//
bool CPIParser::Parse( CScanner* pScanner)
{
return pScanner->GetNextStringMatch((const char*)"?>");
}

//
// Function Name:bool CChardata::Parse( CScanner* pScanner)
// Parse the remained part of Chardata
//
bool CChardata::Parse( CScanner* pScanner)
{
	if (!pScanner->GetNextStringMatch((const char*)"<"))
		return false;
	
	return true;
}



//
//	Function Name:
//		void CAttribute::SetValue(const char* Value);
//	Descriptions:
//  Set the value of the attributes
void CAttribute::SetValue(const char* Value)
{
	CStringValue s(const_cast<char*>(Value), true);
	m_AttributeValue=s;
}

//
// Function Name:
// bool CAttribute::Parse( CScanner* pScanner)
//
bool CAttribute::Parse( CScanner* pScanner)
{
	pScanner->SkipWhiteSpace();
    CToken* Token=	pScanner->GetNextToken();
	CToken::TokenCode tc=Token->GetTokenCode();

	if (tc==CToken::TC_LeftArrow||tc==CToken::TC_ForwardSlash)
	{
		pScanner->BackOneStep();
		return true;
	}

   if (tc!=CToken::TC_Word)
   {
	   return false;
   }

   CWordToken * wordToken =static_cast<CWordToken *>(Token);

   //
   // Get the attribute name
   //
   m_AttributeName.m_pBegin=const_cast<char*>(wordToken->GetTokenString());
   m_AttributeName.m_pEnd=const_cast<char*>(pScanner->GetCurrentPos());

   CToken * eqToken=	pScanner->GetNextToken();
   if (eqToken->GetTokenCode()!=CToken::TC_Equal)
   {
	   return false;
   }
   

    CToken * QuoteToken=	pScanner->GetNextToken();
	if (QuoteToken->GetTokenCode()!=CToken::TC_Quote)
	{
		return false;
	}

	//
	// Get the attribute values.
	//
    m_AttributeValue.m_pBegin=const_cast<char*>(
		static_cast<CQuoteToken *> (QuoteToken)->GetTokenString());
    m_AttributeValue.m_pEnd=const_cast<char*>(pScanner->GetCurrentPos());

	CAttribute* p =new CAttribute(*this);
	p->GetParent()->AddAttribute(p);


   pScanner->SetCurrentPos(m_AttributeValue.m_pEnd+1);

   return true;
}


//
//	Function name:
// bool CStagParser::Parse( CScanner* pScanner)
//
//  Purpose:
//  Parse the XML STag.
//
//		STag		::=    '<' Name (S Attribute)* S? '>'
//
bool CStagParser::Parse( CScanner* pScanner)
{
   CToken * TokLeftArrow=	pScanner->GetNextToken();

   if (TokLeftArrow->GetTokenCode()!=CToken::TC_LeftArrow)
   {
	   return false;
   }

   CToken * TokWord=pScanner->GetNextToken();
   CToken::TokenCode tc=TokWord->GetTokenCode();
   if (tc!=CToken::TC_Word)
   {
	   //
	   // Check if it is a PI or COMMENT
	   //
	   if (tc==CToken::TC_Exclaimation)
	   {
		   tc=pScanner->GetNextToken()->GetTokenCode();
 			if (tc==CToken::TC_Dash)
			{
				tc=pScanner->GetNextToken()->GetTokenCode();
	 			if (tc==CToken::TC_Dash)
				{
					//
					// It is a comments elements.
					//
					CCommentParser cp(this->GetParent());
					if( cp.Parse(pScanner))
					{
						m_bIsComment=true;
						return true;
					}
 				}
			}
	   }
	   else  if (tc==CToken::TC_QuestionMark)
	   {
		   //
		   // Parse it as PI
		   //
		   CPIParser pp(this->GetParent());
		   if(pp.Parse(pScanner))
		   {
			   m_bIsPI=true;
			   return true;
		   }

	   }
	   return false;
   }

   //
   // Remember the start address and the end address of the <name>
   //
   m_StringValue.m_pBegin=const_cast<char*>(TokWord->GetTokenString());
   m_StringValue.m_pEnd=const_cast<char*>(pScanner->GetCurrentPos());

   //
   // Parse the attribute list
   //
   bool bParseAttrList=true;
   CAttribute AttrParser(this->GetParent());

   CToken* TokWhiteOrRightArrow= pScanner->GetNextToken();
   CToken::TokenCode tc_whiteOrRightArrow=TokWhiteOrRightArrow->GetTokenCode();

   while (bParseAttrList)
   {
		if (tc_whiteOrRightArrow==CToken::TC_WhiteSpace)
		{
			if (!AttrParser.Parse(pScanner))
				return false;

   		    tc_whiteOrRightArrow=pScanner->GetNextToken()->GetTokenCode();
		}
		else
		{
			bParseAttrList=false;
		}
   }

   //
   // Parse the '>'
   //
   if (tc_whiteOrRightArrow==CToken::TC_RightArrow)
   {
	   m_bIsEmptyElementTag=false;
	   return true;
   }

   if (tc_whiteOrRightArrow==CToken::TC_ForwardSlash)
   {
	   CToken::TokenCode tc=pScanner->GetNextToken()->GetTokenCode();
	   if (tc==CToken::TC_RightArrow)
	   {
		   m_bIsEmptyElementTag=true;
		   return true;
	   }
  }

  return false;
}

//
// Function name:
// bool CEtagParser::Parse( CScanner* pScanner)
//
// Desc:
// Parse the Etag element
//
bool CEtagParser::Parse( CScanner* pScanner)
{
   CToken * TokLeftArrow=	pScanner->GetNextToken();
   if (TokLeftArrow->GetTokenCode()!=CToken::TC_LeftArrow)
   {
	   return false;
   }

   CToken * TokForward=pScanner->GetNextToken();
   if (TokForward->GetTokenCode()!=CToken::TC_ForwardSlash)
   {
		return false;
   }


   CToken * TokWord=pScanner->GetNextToken();
   if (TokWord->GetTokenCode()!=CToken::TC_Word)
   {
		return false;
   }

   m_StringValue.m_pBegin= const_cast<char*>(TokWord->GetTokenString());
   m_StringValue.m_pEnd=const_cast<char*>(pScanner->GetCurrentPos());

     
   CToken * TokRightArrow=	pScanner->GetNextToken();
   if (TokRightArrow->GetTokenCode()!=CToken::TC_RightArrow)
   {
		return false;
   }

   return true;
}


bool CElement::Parse(CScanner* pScan)
{
	CStagParser StagParser(this);
	CEtagParser EtagParser(this);

	//
	// Parse the begin tag
	//
	if (!StagParser.Parse(pScan))
	{
		return false;
	}

	if (StagParser.IsEmptyElementTag()||StagParser.IsPITag()||StagParser.IsCommentTag())
	{
		m_StringValue=StagParser.GetNameObj();
		if (m_pParent)
		{
			m_pParent->AddChildElement(this);
		}
		return true;
	}
 
	//
	// Parse the content
	//
	CContent contentParser(this);
	if (!contentParser.Parse(pScan))
	{
		return false;
	}

    //
	// Parse the end tag
	//
	if (!EtagParser.Parse(pScan))
	{
		return false;
	}

	if(StagParser.GetNameObj()==EtagParser.GetNameObj())
	{
		m_StringValue=StagParser.GetNameObj();
		//
		// Parse success, add the element
		//
		if (m_pParent)
		{
			m_pParent->AddChildElement(this);
		}
		return true;
	}
	return false;
}

bool CContent::Parse(CScanner* pScan)
{
	//
	// Check the leading contents and decide if we need to 
	//
	const char* pCurrentPos=pScan->GetCurrentPos();
	CToken* TheToken=pScan->GetNextToken();
	CToken::TokenCode tc=TheToken->GetTokenCode();

	bool bChardataParsed=false;

	while (tc!=CToken::TC_Eof&&tc!=CToken::TC_Error)
	{
		switch (tc)
		{
		case CToken::TC_WhiteSpace:
			break; // Get the next token and check
			
		case CToken::TC_Word:
			{
				if (bChardataParsed)
				{
					return false;
				}
				//
				// Keep the start pointer of the CData
				//
				const char* pCurrentPos=TheToken->GetTokenString();
				
				// CData part
				CChardata cdataParser(this->GetParent());
				cdataParser.GetStringObj().m_pBegin=const_cast<char*>(pCurrentPos);
				if (!cdataParser.Parse(pScan))
				{
					return false;
				}

				//
				// Keep the end pointer of the CData
				//
				pScan->BackOneStep();
				cdataParser.GetStringObj().m_pEnd=const_cast<char*>(pScan->GetCurrentPos());
 				m_pParent->SetCharData(cdataParser);

 				bChardataParsed=true;
				break;
			}
		case CToken::TC_LeftArrow:
			{
				//
				// Check the following three conditions
				//
				// <!, </, <abc
				char* pCurrent=(char*)pScan->GetCurrentPos();
				if (CharTypeTable::ChartypeTable [(unsigned char)*pCurrent]==Exclaimation)
				{
					// This is comment
					pScan->SetCurrentPos(pCurrentPos);
					CCommentParser commonParser(this->GetParent());
					if (!commonParser.Parse(pScan))
					{
						return false;
					}
				}
				else if (CharTypeTable::ChartypeTable [(unsigned char)*pCurrent]==ForwardSlash)
				{
					//
					// This is an etag, my job is done.
					pScan->SetCurrentPos(pCurrentPos);
					return true;
				}
				else if (CharTypeTable::ChartypeTable [(unsigned char)*pCurrent]==NormalChar)
				{
					//
					// This is another element, parse it
					//
					pScan->SetCurrentPos(pCurrentPos);
					CElement*pElementParser=new  CElement(this->GetParent());
					if (!pElementParser->Parse(pScan))
					{
						delete pElementParser;
						return false;
					}
				}
				//
				// Otherwise, return false
				//
 				break;
			}
		default:
			return false;
		};
		pCurrentPos=pScan->GetCurrentPos();
	    TheToken=pScan->GetNextToken();
		tc=TheToken->GetTokenCode();
	}

	return false;
}

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//	THE MEMBER FUNCTIONS DEFINITIONS FOR THE CLASS CElement
//

//
// Function Name: 
// void CElement::Clean()
//
// Comment: Clean up all the subelements and sibling elements
//
void CElement::Clean()
{

	CElement*p=m_pFirstChild;
	while (p)
	{
		CElement*tmp=p;
		p=p->m_pNextSibling;
		delete tmp;

		if (p==m_pFirstChild)
		{
			break;
		}
	}
	//
	// Cleanup the attribute list
	//
	vector <CAttribute*>::iterator Abegin=m_AttributeList.begin();
	vector <CAttribute*>::iterator Aend=m_AttributeList.end();
	while (Abegin!=Aend)
	{
		delete *Abegin;
		Abegin++;
	}
	m_AttributeList.clear();


}

//
// Function Name:void CElement::
//	 AddChildElement(CStringValue& sv, CElement*p);
//
// Add the child elements of the current elements
//
void CElement::AddChildElement( CElement*p)
{
	//
	// Configure the sibling chain
	//
	if (m_pFirstChild)
	{
		p->m_pNextSibling=m_pFirstChild;
		p->m_pPrevSibling=m_pFirstChild->m_pPrevSibling;
		m_pFirstChild->m_pPrevSibling->m_pNextSibling=p;
		m_pFirstChild->m_pPrevSibling=p;
	}
	else
	{
		m_pFirstChild=p;
		m_pFirstChild->m_pPrevSibling=m_pFirstChild->m_pNextSibling=m_pFirstChild;
	}
}


//
// Function Name:void CElement::DeleteChild(CElement*p)
//
// Delete the child referenced by CElement*
//
bool CElement::DeleteChild (CElement*pRMV)
{
	if (!m_pFirstChild) return false;
	CElement*p=m_pFirstChild;
	bool bEndofLoop=false;
	while (p!=pRMV && !bEndofLoop)
	{
		p=p->m_pNextSibling;
		if (p==m_pFirstChild)
		{
			bEndofLoop=true;
		}
	}

	if (p==pRMV)
	{
		CElement* pFirstCandidate=m_pFirstChild;
		if (pRMV==m_pFirstChild)
		{
			pFirstCandidate=m_pFirstChild->m_pNextSibling;
		}
		bool bSingleChild=(m_pFirstChild->m_pPrevSibling==m_pFirstChild);
		pRMV->m_pPrevSibling->m_pNextSibling=pRMV->m_pNextSibling;
		pRMV->m_pNextSibling->m_pPrevSibling=pRMV->m_pPrevSibling;
		delete pRMV;
		if (bSingleChild) pFirstCandidate=NULL;

		m_pFirstChild=pFirstCandidate;
		
		return true;
	}
	return false;
}

//
// Function Name:void CElement::DeleteChild(CElementIterator& it)
//
// Delete the child referenced by CElementIterator& it
//
bool CElement::DeleteChild (CElementIterator& it)
{
	if (!it.IsValid()) return false;

	bool ret=DeleteChild(it.GetElementPtr());
	if (ret) it.Invalidate();

 	return ret;
}


//
// Function Name:
// std::ostream& operator<< (std::ostream& ,CElement const&);
//
// Comments:
// Output the XML stream from the current element
//
std::ostream& operator<< (std::ostream&os ,CElement &ElementParser)
{
	//
	// Begin Tag
	//
	os<<endl<<'<'<<ElementParser.GetName();
	
	//
	// Cleanup the attribute list
	//
	vector <CAttribute*>::iterator Abegin=ElementParser.m_AttributeList.begin();
	vector <CAttribute*>::iterator Aend=ElementParser.m_AttributeList.end();
	while (Abegin!=Aend)
	{
		os <<' '<<**Abegin;
		Abegin++;
	}

	os<<'>';

	//
	// Output the data part
	//
	os<<ElementParser.m_CharData;

 
 	//
	// Walk through the children chains and output each child
	//
	CElement* pElement=ElementParser.m_pFirstChild;
	bool bIsLastItem=false;
	while (!bIsLastItem &&pElement)
	{
		os<<*pElement;
		pElement=pElement->m_pNextSibling;

		if (!pElement||pElement==ElementParser.m_pFirstChild)
			bIsLastItem=true;
	}
	//
	// End tag
	//
	os<<endl<<"</"<<ElementParser.GetName()<<'>';
 	return os;
}


CElementIterator CElement::GetFirstChild()
{
	CElementIterator iterator(m_pFirstChild);
	return iterator;
}


//
// ================================================================================
//
// Begin the attributes functions of the CElement
//
// 

 //
 //	Function name:int CElementIterator::GetAttributeCount()
 //  Purpose:
 // Get the number of the attributes of the element represented by the iterator
//
int CElement::GetAttributeCount()
{
	if (!IsValid()) return 0;
	return m_AttributeList.size();
}

bool CElement::GetAttributePairByIndex(int index, vector<char>&attName, vector<char>&attValue)
{
	if (index>=0 &&index<GetAttributeCount())
	{
		CAttribute* p=m_AttributeList[index];

		attName.resize(p->GetAttrNameObj().GetSize());
		std::copy(p->GetAttrNameObj().m_pBegin,p->GetAttrNameObj().m_pEnd,attName.begin());

		attValue.resize(p->GetAttrValueObj().GetSize());
		std::copy(p->GetAttrValueObj().m_pBegin,p->GetAttrValueObj().m_pEnd,attValue.begin());

 		return true;
	}
	return false;
}

bool CElement::GetAttribute(const char*name, vector<char>&attValue)
{
	int size=GetAttributeCount();
	
	for (int i=0; i<size;i++)
	{
		CAttribute* p=m_AttributeList[i];
		if (const_cast<CStringValue&>(p->GetAttrNameObj()).Equal(name))
		{
 				attValue.resize(p->GetAttrValueObj().GetSize());
				std::copy(p->GetAttrValueObj().m_pBegin,
						  p->GetAttrValueObj().m_pEnd,
						  attValue.begin());
 			return true;

		}
	}
	return false;
}

//
// Function Name:
// 	void CElement::AddAttributePair(const char*Name, const char*Value);
//
// Comments:
// Add the attributes value pair to the Element
//
void CElement::AddAttributePair(const char*Name, const char*Value)
{
	//
	// Create two string (enforece the internal buffer)
	//
	CStringValue strName(const_cast<char*>(Name),true);
	CStringValue strValue(const_cast<char*>(Value),true);
	CAttribute* p=new CAttribute(strName,strValue,this);
	AddAttribute(p);
}

//
// Function Name:
// 	bool CElement::ModifyAttribute(const char*Name, const char*Value)
//
// Comments:
//
// Update the attribute value represented by 'Name'. 
// Return false if the attribute was not found.
//
bool CElement::ModifyAttribute(const char*Name, const char*Value)
{
	int size=GetAttributeCount();
	
	for (int i=0; i<size;i++)
	{
		CAttribute* p=m_AttributeList[i];

		if (const_cast<CStringValue&>(p->GetAttrNameObj()).Equal(Name))
			{
 				p->SetValue(Value);
 				return true;
			}
	}
	return false;
}

//
// Function Name:
// 	bool CElement::ModifyAttributeByIndex((int index, const char*Value)
//
// Comments:
//
// Update the attribute value represented by 'index'. 
// Return false if the attribute was not found.
//
bool CElement::ModifyAttributeByIndex(int index, const char*Value)
{
	int size=GetAttributeCount();
	
	for (int i=0; i<size;i++)
	{
		if(i==index)
		{
			CAttribute* p=m_AttributeList[i];
 			p->SetValue(Value);
 			return true;
		}
	}
	return false;
}

//
// Function Name:
// 	bool CElement::DeleteAttributeByName(const char*Name)
//
// Comments:
bool CElement::DeleteAttributeByName(const char*name)
{
	int size=GetAttributeCount();
	vector<CAttribute*>::iterator begin=m_AttributeList.begin();
	vector<CAttribute*>::iterator end=m_AttributeList.end();
	while (begin!=end)
	{
		CAttribute*p=(*begin);
		if (const_cast<CStringValue&>(p->GetAttrNameObj()).Equal(name))
		{
				delete p;
				m_AttributeList.erase(begin);
				return true;
		}
 		begin++;
	}
	return false;
}

//
// Function Name:
// 	bool CElement::DeleteAttributeByIndex(int index)
//
// Comments:
bool CElement::DeleteAttributeByIndex(int Index)
{
	int size=GetAttributeCount();
	vector<CAttribute*>::iterator begin=m_AttributeList.begin();
	vector<CAttribute*>::iterator end=m_AttributeList.end();
	int index=0;
	while (begin!=end)
	{
		if (index==Index)
		{
			CAttribute*p=(*begin);
			delete p;
			m_AttributeList.erase(begin);
			return true;
		}
		index++;
		begin++;
	}
	return false;
}













//
// End of Attributes functions
// ================================================================================




//
//	Function name: 	bool CElement::GetCharData(vector<char>&ret);
//
//  Purpose:
//  Return the string value to the 'ret'. Return false if the Chardata is not
//  available.
//
bool CElement::GetCharData(vector<char>&ret)
{
	if (m_CharData)
	{
		ret.resize(m_CharData.GetStringObj().m_pEnd-m_CharData.GetStringObj().m_pBegin);
		std::copy(m_CharData.GetStringObj().m_pBegin,m_CharData.GetStringObj().m_pEnd,ret.begin());
		ret.push_back('\0');
		return true;
	}
	return false;
}


void CElement::SetCharData(CStringValue& s)
{
	m_CharData.SetStringObj(s);
	m_CharData.SetParent(this);
	m_CharData.SetValid();
}


std::ostream& operator<< (std::ostream&os ,CElement const&ElementParser)
{
	return os<<const_cast<CElement&>(ElementParser);
}


CElement* CElement::CreateNewElement(const char* Name, CElement*parent)
{
	CStringValue strValue(const_cast<char*>(Name),true);

	CElement* p= new CElement(parent);
	p->SetName(strValue);
	return p;
}


////////////////////////////////////////////////////////////////////////////////////////////////////
// 
//			THE MEMBER FUNCTIONS OF THE 
//
//			class CElementIterator
//
//

//
// Copy constructors and operator =
//

CElementIterator::CElementIterator (CElement* pObj):m_bValid(false)
{
	if (pObj)
	{
		m_pCurrentElement=pObj;
		m_bValid=true;
	}
}

CElementIterator::CElementIterator(CElementIterator const &obj):
m_bValid(obj.m_bValid),m_pCurrentElement(obj.m_pCurrentElement)
{}

CElementIterator& CElementIterator::operator= (CElementIterator const&obj)
{
	if (this!=&obj)
	{
		m_bValid=obj.m_bValid;
		m_pCurrentElement=obj.m_pCurrentElement;
	}
	return *this;
}


//
// Function Name: CElementIterator& CElementIterator::operator++ ()
//
// Comments:
//
// move the CElementIterator to the next Elements that share the same
// root.
//
CElementIterator& CElementIterator::operator++ ()
{
	if(!IsValid())
	{
		return *this;
 	}

	if (m_pCurrentElement->m_pNextSibling && 
		m_pCurrentElement->m_pNextSibling!=m_pCurrentElement->GetParent()->m_pFirstChild)
	{
		m_pCurrentElement=m_pCurrentElement->m_pNextSibling;
	}
	else
	{
		m_bValid=false;
	}
	return *this;
}

//
// Function Name: void CElementIterator::Delete ()
//
// Comments:
//
// delete the associated element. After deletion, the element iterator will become invalidated.
// 
void CElementIterator::Delete ()
{
	if (IsValid())
	{
		CElement* pParent=	m_pCurrentElement->GetParent();
		if (!pParent)
		{
 			// This is the root, just delete it
 			delete m_pCurrentElement;
		}
		else
		{
			// Should call something like pParent->DeleteChild(m_pCurrentElement);

		}

		m_bValid=false;
		m_pCurrentElement=NULL;
	}

}
