// Token.cpp: implementation of the CToken class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Token.h"
using namespace std;

CharType CharTypeTable::ChartypeTable[256]; 

//
//	Function name:void CharTypeArray::InitCharType()
//  Purpose:
//		Fill character type table.
//
void CharTypeTable::InitCharType()
{
	for (int i=0;i<256;i++) 
	{
		char ch= static_cast<char>(i);
		if (ch>='a'&&ch<='z'||
			ch>='A'&&ch<='Z'||
			ch>='0'&&ch<='9'||
			ch=='@'||
			ch=='$'||
			ch=='^'||
			ch=='&'||
			ch=='*'||
			ch=='('||
			ch==')'||
			ch=='_'||
			ch=='+'||
			ch==':'||
			ch=='{'||
			ch=='}'||
			ch=='['||
			ch==']'||
			ch=='.'||
			ch=='~'||
			ch=='\\')
		{
			CharTypeTable::ChartypeTable[i]=NormalChar;
		}
		else
		{
			CharTypeTable::ChartypeTable[i]=ErrorChar;
		}
	}

	CharTypeTable::ChartypeTable[' ']=WhiteSpace;
	CharTypeTable::ChartypeTable['\t']=WhiteSpace;
	CharTypeTable::ChartypeTable['\n']=WhiteSpace;
	CharTypeTable::ChartypeTable['\r']=WhiteSpace;
	CharTypeTable::ChartypeTable['<']=LeftArrow;
	CharTypeTable::ChartypeTable['>']=RightArrow;
	CharTypeTable::ChartypeTable['?']=QuestionMark;
	CharTypeTable::ChartypeTable['\"']=Quote;
	CharTypeTable::ChartypeTable['/']=ForwardSlash;
	CharTypeTable::ChartypeTable['-']=Dash;
	CharTypeTable::ChartypeTable['!']=Exclaimation;
	CharTypeTable::ChartypeTable['=']=Equal;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
// MEMBER FUNCTIONS FOR THE CLASS CStringValue
// MEMBER FUNCTIONS FOR THE CLASS CStringValue
// MEMBER FUNCTIONS FOR THE CLASS CStringValue
// MEMBER FUNCTIONS FOR THE CLASS CStringValue
// MEMBER FUNCTIONS FOR THE CLASS CStringValue
//
//
//

//
// Function Name: CStringValue Destructor
//
//
CStringValue::~CStringValue()
{
	if (m_bUseMyBuffer) delete []m_Buffer;
}

//
// Function Name: CStringValue Constructor
//
CStringValue::CStringValue(char*pBegin, char*pEnd, bool bUseInternalBuffer)
{
	m_bUseMyBuffer=bUseInternalBuffer;
	if (m_bUseMyBuffer)
	{
		m_Buffer=new char[pEnd-pBegin];
		memcpy(m_Buffer,pBegin,pEnd-pBegin);
		m_pBegin=m_Buffer;
		m_pEnd=m_Buffer+(pEnd-pBegin);
	}
	else
	{
		m_pBegin=pBegin;
		m_pEnd=pEnd;
	}
}


//
// Function Name: CStringValue Constructor
//
CStringValue::CStringValue(char*pBegin, bool bUseInternalBuffer)
{
	m_bUseMyBuffer=bUseInternalBuffer;
	int stringSize=strlen(pBegin);
	if (m_bUseMyBuffer)
	{
		m_Buffer=new char[stringSize];
		std::copy(pBegin,pBegin+stringSize,m_Buffer);
		m_pBegin=m_Buffer;
		m_pEnd=m_Buffer+stringSize;
	}
	else
	{
		m_pBegin=pBegin;
		m_pEnd=pBegin+stringSize;
	}
}


//
// Function Name: CStringValue copy constructor
//
//
CStringValue::CStringValue (CStringValue const &obj)
{
 		if (obj.m_bUseMyBuffer)
		{
			m_Buffer=new char[obj.m_pEnd-obj.m_pBegin];
			memcpy(m_Buffer,obj.m_Buffer,obj.m_pEnd-obj.m_pBegin);
			m_pBegin=m_Buffer;
			m_pEnd=m_Buffer+(obj.m_pEnd-obj.m_pBegin);
			m_bUseMyBuffer=true;
		}
		else
		{
			m_pBegin=obj.m_pBegin;
			m_pEnd=obj.m_pEnd;
			m_bUseMyBuffer=false;
		}
}


//
// Function Name: CStringValue operator=
//
//
CStringValue& CStringValue::operator= (CStringValue const &obj)
{
	if (this!=&obj)
	{
		if (m_bUseMyBuffer) 
		{
			delete []m_Buffer;
		}

		if (obj.m_bUseMyBuffer)
		{
			m_Buffer=new char[obj.m_pEnd-obj.m_pBegin];
			memcpy(m_Buffer,obj.m_Buffer,obj.m_pEnd-obj.m_pBegin);
			m_pBegin=m_Buffer;
			m_pEnd=m_Buffer+(obj.m_pEnd-obj.m_pBegin);
			m_bUseMyBuffer=true;
		}
		else
		{
			m_pBegin=obj.m_pBegin;
			m_pEnd=obj.m_pEnd;
			m_bUseMyBuffer=false;
		}
	}
	return *this;
}





//
// Function Name: 
// 	friend bool operator==(CStringValue const& v1, CStringValue const& v2);
// Compare the two strings values
//
bool operator==(CStringValue const& v1, CStringValue const& v2)
{
	if (v1.m_pEnd-v1.m_pBegin==v2.m_pEnd-v2.m_pBegin)
	{
		return std::equal(v1.m_pBegin,v1.m_pEnd,v2.m_pBegin);
	}
	return false;
}

bool operator!=(CStringValue const& v1, CStringValue const& v2)
{
	return !(v1==v2);
}

//
// Member functions of CStringValue
//
std::ostream& operator <<(std::ostream& os, CStringValue const&obj)
{
	char*p=obj.m_pBegin;
  while(p<obj.m_pEnd)
  {
	  os<<*p;
	  p++;
  }
  return os;
}

bool CStringValue::Equal(const char*str)
{
	if (GetSize()!=strlen(str))
		return false;
	return std::equal(m_pBegin,m_pEnd,str);
}

//
// END END END
// END END END
// END END END
// MEMBER FUNCTIONS FOR THE CLASS CStringValue
//
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////


//
// Function name: void CToken::AcquireToken 
//
// Comments:
// The basic acquiretoken action
//
bool CToken::AcquireToken(	const char* pCurrent,
							const char* pEndPos,  // The end postion is the last legal position plus 1
							char* &postPostition) 
{
	m_pToken=pCurrent;
	postPostition=const_cast<char*>(m_pToken)+1;
	return true;
}


ostream& operator <<(ostream& os, CToken const& obj)
{
	os<<"<";
	if(typeid(obj)==typeid(CQuoteToken))
	{
 		os<<"CQuoteToken";
 	}
	else if (typeid(obj)==typeid(CLeftArrowToken))
	{
 		os<<"CLeftArrowToken";
 	}
	else if (typeid(obj)==typeid(CRightArrowToken))
	{
 		os<<"CRightArrowToken";
 	}
	else if (typeid(obj)==typeid(CQuestionMarkToken))
	{
 		os<<"CQuestionMarkToken";
 	}
	else if (typeid(obj)==typeid(CForwardSlashToken))
	{
 		os<<"CForwardSlashToken";
 	}
	else if (typeid(obj)==typeid(CQuoteToken))
	{
 		os<<"CQuoteToken";
 	}
	else if (typeid(obj)==typeid(CDashToken))
	{
 		os<<"CDashToken";
 	}
	else if (typeid(obj)==typeid(CExclaimationToken))
	{
 		os<<"CExclaimationToken";
 	}
	else if (typeid(obj)==typeid(CWordToken))
	{
 		os<<"CWordToken";
 	}
	return os<<">";
}



bool CQuoteToken::AcquireToken(const char* pCurrent,
							  const char* pEndPos,  // The end postion is the last legal position plus 1
							  char* &postPostition)
{
	char* pScan=const_cast<char*>(pCurrent);

	char* pFind=std::find(const_cast<char*>(pCurrent),const_cast<char*>(pEndPos),'\"');
	if (pFind==pEndPos)
	{
		return false;
	}
	postPostition=pFind;
	m_pToken=pCurrent;
	return true;
}
 
bool CWordToken::AcquireToken(const char* pCurrent,
							  const char* pEndPos,  // The end postion is the last legal position plus 1
							  char* &postPostition)
{
	char* pScan=const_cast<char*>(pCurrent);
 	while (pScan<pEndPos)
	{
		if (CharTypeTable::ChartypeTable[(unsigned char)*pScan]==NormalChar) pScan++;
		else
		{
			m_pToken=pCurrent;
			break;
		}
	}
		
	postPostition=pScan;
	return true;
}

bool CWhiteSpace::AcquireToken(const char* pCurrent,
							  const char* pEndPos,  // The end postion is the last legal position plus 1
							  char* &postPostition)
{
	char* pScan=const_cast<char*>(pCurrent);
 	while (pScan<pEndPos)
	{
		if (CharTypeTable::ChartypeTable[(unsigned char)*pScan]==WhiteSpace) pScan++;
		else
		{
			m_pToken=pCurrent;
			break;
		}
	}
	
	postPostition=pScan;
	return true;
}


//
//	Function name: const CToken* CScanner::GetNextToken()
//  Purpose:
//
//  Scan the char stream referenced by the internal pointer and
//  return a token. The function will mark the end of token position
//  with null terminator.
//
// Return:
// The function return its internal token object (so don't remove scanner objcet)
// when using the the returned token object.).
//
// The function returns NULL when NoToken was found in the string stream.
//
CToken* CScanner::GetNextToken()
{
	if (m_pCurrentPos>=m_EndPos) return &m_EofToken;

	char* pCurrentPosition=NULL;
   while (m_pCurrentPos<m_EndPos)
   {
    switch (CharTypeTable::ChartypeTable[(unsigned char)*m_pCurrentPos])
	{
	case ErrorChar:
		return &m_ErrorToken;
	case NormalChar:
		m_WordToken.AcquireToken(m_pCurrentPos,m_EndPos,pCurrentPosition);
 		m_pCurrentPos=pCurrentPosition;
		return &m_WordToken;
	case LeftArrow:
		m_LeftArrowToken.AcquireToken(m_pCurrentPos,m_EndPos,pCurrentPosition);
		m_pCurrentPos=pCurrentPosition;
		return &m_LeftArrowToken;
	case RightArrow:
		m_RightArrowToken.AcquireToken(m_pCurrentPos,m_EndPos,pCurrentPosition);
		m_pCurrentPos=pCurrentPosition;
		return &m_RightArrowToken;

	case QuestionMark:
		m_QuestionMarkToken.AcquireToken(m_pCurrentPos,m_EndPos,pCurrentPosition);
		m_pCurrentPos=pCurrentPosition;
		return &m_QuestionMarkToken;

	case ForwardSlash:
		m_ForwardSlashToken.AcquireToken(m_pCurrentPos,m_EndPos,pCurrentPosition);
		m_pCurrentPos=pCurrentPosition;
		return &m_ForwardSlashToken;

	case Dash:
		m_DashToken.AcquireToken(m_pCurrentPos,m_EndPos,pCurrentPosition);
		m_pCurrentPos=pCurrentPosition;
		return &m_DashToken;

	case Exclaimation:
		m_ExclaimationToken.AcquireToken(m_pCurrentPos,m_EndPos,pCurrentPosition);
		m_pCurrentPos=pCurrentPosition;
		return &m_ExclaimationToken;

	case Quote:
		if (m_QuoteToken.AcquireToken(m_pCurrentPos+1,m_EndPos,pCurrentPosition))
		{
			m_pCurrentPos=pCurrentPosition;
			return &m_QuoteToken;
		}
		else
		{
			return &m_ErrorToken;
		}
	case WhiteSpace:
		{
			m_WhiteSpaceToken.AcquireToken(m_pCurrentPos,m_EndPos,pCurrentPosition);
 			m_pCurrentPos=pCurrentPosition;
			return &m_WhiteSpaceToken;
		}
	case Equal:
		{
			m_EqualToken.AcquireToken(m_pCurrentPos,m_EndPos,pCurrentPosition);
 			m_pCurrentPos=pCurrentPosition;
			return &m_EqualToken;
		}


	default:
		//No default, I would be in trouble if the program ran in this branch!!!
	   return &m_ErrorToken;
		break;
	};
   }
   return &m_EofToken;
}

/********************************************************************
	created:	22:5:2005   12:36
	function Name: 	CScanner::SkipWhiteSpace
	purpose:	
*********************************************************************/
void CScanner::SkipWhiteSpace()
{
	while (m_pCurrentPos<m_EndPos &&
		CharTypeTable::ChartypeTable[(unsigned char)*m_pCurrentPos]==WhiteSpace)
		m_pCurrentPos++;
}


//
//	Function name: bool CScanner::GetNextStringMatch (const char* match)
//  Purpose:
//
// The function scan the buffer from the m_pCurrentPtr and tries to find the
// matched string for the <Key>. If the function successes, the m_pCurrentPtr
// should stop to the next char position after matched string.
//
// otherwise the function returns false.
//
// This function is used to quickly parse the comments and PI only.
//
bool CScanner::GetNextStringMatch (const char* Key)
{
	if (m_pCurrentPos>=m_EndPos) return false;

	int KeySize=strlen(Key);

	char* pFind=std::search(m_pCurrentPos,
									m_EndPos,
									Key,
									Key+KeySize);

	if (pFind== m_EndPos)
	{
		return false;
	}
	m_pCurrentPos=pFind+KeySize;
	return true;
}


//
// Function Name:
//