// Token.h: interface for the CToken class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_TOKEN_H__EB8568FF_A025_42FD_A377_3C3035FB4E75__INCLUDED_)
#define AFX_TOKEN_H__EB8568FF_A025_42FD_A377_3C3035FB4E75__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

enum CharType
{
	NormalChar,
	ErrorChar,
	Eof,
	LeftArrow,
	RightArrow,
	QuestionMark,
	Quote,
	ForwardSlash,
	Dash,
	WhiteSpace,
	Exclaimation,
	Equal
};

struct CharTypeTable
{
	static  CharType ChartypeTable[256];
	static void InitCharType();
};

//
// Define structure of CStringValue
//
class CStringValue
{
public:
	char* m_pBegin;
	char* m_pEnd;



	//
	// The default constructor is only used for the parsing stage,
	// which assumes that the external buffer is provided.(for efficiency purpose)
	//
	CStringValue():m_pBegin(NULL),m_pEnd(NULL),m_bUseMyBuffer(false),m_Buffer(NULL){}
	~CStringValue();
	CStringValue(char*pBegin, char*pEnd, bool bUseInternalBuffer);
	CStringValue(char*pBegin, bool bUseInternalBuffer);//For null terminated string
	CStringValue (CStringValue const &);
	CStringValue& operator= (CStringValue const &);

	inline operator bool () const {return m_pBegin!=NULL&&m_pEnd!=NULL;}
	inline int GetSize() const {return (*this)?m_pEnd-m_pBegin:0;}
	bool Equal(const char*str);
 	
	//
	// Warining: This function is only used in the parsing stage.
	//
	
	
	
	friend bool operator==(CStringValue const& v1, CStringValue const& v2);
	friend bool operator!=(CStringValue const& v1, CStringValue const& v2);
	friend std::ostream& operator <<(std::ostream&, CStringValue const&);


private:
	bool m_bUseMyBuffer;
	char* m_Buffer;

};

class CToken  
{
public:
	enum TokenCode
	{
		TC_Word,
		TC_Error,
		TC_Eof,
		TC_LeftArrow,
		TC_RightArrow,
		TC_QuestionMark,
		TC_Quote,
		TC_ForwardSlash,
		TC_Dash,
		TC_WhiteSpace,
		TC_Exclaimation,
		TC_Equal
	};

	CToken():m_pToken(NULL){}
	virtual ~CToken(){};
	virtual bool AcquireToken( const char* pCurrent,
							   const char* pEndPos,  
							   char* &postPostition);
	inline const char* GetTokenString() const {return m_pToken;}
	inline TokenCode  GetTokenCode() const{return m_TokenCode;}
	friend std::ostream& operator <<(std::ostream& os, CToken const& obj);



protected:
	const char* m_pToken;
	TokenCode  m_TokenCode;

private:

	//
	// Hide the operator= and copy constuctor to prevent the subclass from
	// copying.
	//
	CToken& operator= (CToken const&);
	CToken (CToken const &);
};


//
// Token of "<"
//
class CLeftArrowToken:public CToken
{
public:
	CLeftArrowToken()
	{
		m_TokenCode=TC_LeftArrow;
	}
};

//
// Token of ">"
//
class CRightArrowToken:public CToken
{
public:
	CRightArrowToken() 
	{
		m_TokenCode=TC_RightArrow;
	}
};

//
// Token of "\"
//
class CForwardSlashToken:public CToken
{
public:
	CForwardSlashToken() 
	{
		m_TokenCode=TC_ForwardSlash;
	}
};

//
// Token of "?"
//
class CQuestionMarkToken:public CToken
{
public:
	CQuestionMarkToken() 
	{
		m_TokenCode=TC_QuestionMark;
	}
};

//
// Token of "\""
//
class CQuoteToken:public CToken
{public:

	CQuoteToken()
	{
		m_TokenCode=TC_Quote;
	}

	virtual bool AcquireToken(const char* pCurrent,
							  const char* pEndPos,  // The end postion is the last legal position plus 1
							  char* &postPostition);
};

//
// Token of "Dash"
//
class CDashToken:public CToken
{
public:
	CDashToken()
	{
		m_TokenCode=TC_Dash;
	}
};

//
// Token of "Exclaimation"
//
class CExclaimationToken:public CToken
{
public:
	CExclaimationToken()
	{
		m_TokenCode=TC_Exclaimation;
	}
};

//
// Eof Token
//
class CEofToken:public CToken
{
public:
	CEofToken()
	{
		m_TokenCode=TC_Eof;
	}
};

//
// Token of "CEqualToken"
//
class CEqualToken:public CToken
{
public:
	CEqualToken()
	{
		m_TokenCode=TC_Equal;
	}
};

//
// Error Token
//
class CErrorToken:public CToken
{
public:
	CErrorToken()
	{
		m_TokenCode=TC_Error;
	}
};

 
//
// Words Token
//
class CWordToken:public CToken
{public:
	CWordToken()
	{
		m_TokenCode=TC_Word;

	}
	virtual bool AcquireToken(const char* pCurrent,
							  const char* pEndPos,  // The end postion is the last legal position plus 1
							  char* &postPostition);
};

//
// WhiteSpace Token
//
class CWhiteSpace:public CToken
{public:
	CWhiteSpace()
	{
		m_TokenCode=TC_WhiteSpace;
	}
	virtual bool AcquireToken(const char* pCurrent,
							  const char* pEndPos,  // The end postion is the last legal position plus 1
							  char* &postPostition);
};


//
// Class name:CScanner
//
// Comments:
// The CScanner runs in the server concept. The GetToken function
// is its center of support.
class CScanner
{
public:
	CScanner(char* pBegin, char* end):
									m_pStartPos(pBegin),
									m_EndPos(end),
									m_pCurrentPos(pBegin)
	{}

    
	bool GetNextStringMatch(const char* pattern);								

	CToken* GetNextToken();
	void SkipWhiteSpace();
 	inline const char* GetCurrentPos() const {return m_pCurrentPos;}
 	inline const char* GetEndPos() const {return m_EndPos;}
	inline void SetCurrentPos(char*p){m_pCurrentPos=p;}
	inline void SetCurrentPos(const char*p){m_pCurrentPos=const_cast<char*>(p);}
 	inline void BackOneStep() {m_pCurrentPos--;	}
	inline void ForwardOneStep() {m_pCurrentPos++;	}
	
private:
	CScanner();								// Hide the default constructor
	CScanner(CScanner const&);				// Hide the copy constructor
	CScanner& operator= (CScanner const&);  // Hide the operator =

	CLeftArrowToken		m_LeftArrowToken;
	CRightArrowToken	m_RightArrowToken;
	CForwardSlashToken  m_ForwardSlashToken;
	CQuestionMarkToken  m_QuestionMarkToken;
	CDashToken			m_DashToken;
	CExclaimationToken  m_ExclaimationToken;
	CQuoteToken			m_QuoteToken;
	CWordToken			m_WordToken;
	CWhiteSpace         m_WhiteSpaceToken;
	CEofToken			m_EofToken;
	CErrorToken			m_ErrorToken;
	CEqualToken			m_EqualToken;


	char* m_pStartPos;
	char* m_pCurrentPos;
	char* m_EndPos;
};

#endif // !defined(AFX_TOKEN_H__EB8568FF_A025_42FD_A377_3C3035FB4E75__INCLUDED_)
