// XmlConf.cpp: implementation of the CXmlConf class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Token.h"
#include "Parser.h"
#include "XmlConf.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
 
CXmlConf::~CXmlConf()
{
	if (m_pScanner) delete m_pScanner;
	if (m_pRoot) delete m_pRoot;
	if (m_pBuffer) m_pBuffer->Delete();
}

//
// Constructor: CXmlConf::CXmlConf(const char* fileName)
//
// Comments:
// Constructor by Reading the xml file from 'fileName',
//
CXmlConf::CXmlConf(const char* fileName):
		m_bValid(false),
		m_pRoot(NULL),
		m_pBuffer(NULL),
		m_pScanner(NULL)
{
   FILE * pFile=fopen(fileName,"r");
   if (!pFile)
	   return;

   fseek (pFile, 0, SEEK_END);
   unsigned int size=ftell (pFile);
   fseek (pFile, 0, SEEK_SET);

   char*pBuffer= new char[size];

   unsigned int bytesRead=0;
   while (bytesRead<size)
   {
	   unsigned int blockSize=fread(pBuffer+bytesRead,1,0x1000,pFile);
	   if (blockSize<0)
	   {
		   fclose(pFile);
 		   delete []pBuffer;
 		   return;
	   }
	   else if (blockSize==0)
	   {
		   break;
	   }

	   bytesRead+=blockSize;
   }
   fclose(pFile);

   if (!Parse(pBuffer,bytesRead))
   {
	   delete[]pBuffer;
   }
   else
   {
	   //
	   // Create the shared memory
	   //
	   m_pBuffer=new SharedBuffer;
	   m_pBuffer->pBuffer=pBuffer;
	   m_pBuffer->refcount=1;
	   m_pBuffer->bufferSize=bytesRead;

	   m_bValid=true;
   }
}

//
// Constructor: CXmlConf::CXmlConf(char* Buffer, bool bAssigned); 
//
// Comments:
// Constructor by getting the input 'Buffer'
//
// If bAssigned is true: the buffer is assigned, we don't need to allocate the internal buffer
// false otherwise.
//
CXmlConf::CXmlConf(char* begin, char* end):
		m_bValid(false),
		m_pRoot(NULL),
		m_pBuffer(NULL),
		m_pScanner(NULL)
{
   int buffersize=end-begin;
   char*pBuffer= new char[buffersize];
   std::copy(begin,end,pBuffer);

   if (!Parse(pBuffer,buffersize))
   {
	   delete[]pBuffer;
   }
   else
   {
	   //
	   // Create the shared memory
	   //
	   m_pBuffer=new SharedBuffer;
	   m_pBuffer->pBuffer=pBuffer;
	   m_pBuffer->refcount=1;
	   m_pBuffer->bufferSize=buffersize;

	   m_bValid=true;
   }

}

bool CXmlConf::Parse(char* pBuffer, int buffersize)
{
   m_pScanner= new CScanner(pBuffer,pBuffer+buffersize);
   m_pRoot=new CElement(NULL);
   m_pScanner->SkipWhiteSpace();

   CPIParser piParser(NULL);
   if (!piParser.Parse(m_pScanner)) 
   {
	   delete []pBuffer;
	   delete m_pRoot;
	   delete m_pScanner;
	   m_pRoot=NULL;
	   m_pScanner=NULL;
	   return false;
   }
   m_pScanner->SkipWhiteSpace();


   if (!m_pRoot->Parse(m_pScanner))
   {
	   delete []pBuffer;
	   delete m_pRoot;
	   delete m_pScanner;
	   m_pRoot=NULL;
	   m_pScanner=NULL;
	   return false;
   }
   return true;
}


CXmlConf::CXmlConf(CXmlConf const&obj):
		m_bValid(false),
		m_pRoot(NULL),
		m_pBuffer(NULL),
		m_pScanner(NULL)
{
	if (obj)
	{
	   if (Parse(obj.m_pBuffer->pBuffer,obj.m_pBuffer->bufferSize))
	   {
 			m_pBuffer=obj.m_pBuffer;
			m_pBuffer->refcount++;	
			m_bValid=true;
	   }
	}
}

		
CXmlConf& CXmlConf::operator=(CXmlConf const&obj)
{
	if (this!=&obj)
	{
		//
		// Cleanup the existing stuff
		//
		if (m_pScanner) delete m_pScanner;
		if (m_pRoot) delete m_pRoot;
		if (m_pBuffer) m_pBuffer->Delete();
		
		if (obj)
		{
		   if (Parse(obj.m_pBuffer->pBuffer,obj.m_pBuffer->bufferSize))
		   {
 				m_pBuffer=obj.m_pBuffer;
				m_pBuffer->refcount++;	
				m_bValid=true;
		   }
		}
	}
	return *this;
}


ostream& operator<<(ostream&os, CXmlConf const& obj)
{
	os<<"<?xml version=\"1.0\" encoding=\"UTF-8\" ?>";
	if (obj&&obj.GetRoot())
		os<<*obj.GetRoot()<<endl;
	return os;
}

